"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Constraints = exports.DependencyType = void 0;

var _core = require("@yarnpkg/core");

var _fslib = require("@yarnpkg/fslib");

var _tauProlog = _interopRequireDefault(require("tau-prolog"));

var _tauModule = require("./tauModule");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncIterator(iterable) { var method; if (typeof Symbol !== "undefined") { if (Symbol.asyncIterator) { method = iterable[Symbol.asyncIterator]; if (method != null) return method.call(iterable); } if (Symbol.iterator) { method = iterable[Symbol.iterator]; if (method != null) return method.call(iterable); } } throw new TypeError("Object is not async iterable"); }

function _awaitAsyncGenerator(value) { return new _AwaitValue(value); }

function _wrapAsyncGenerator(fn) { return function () { return new _AsyncGenerator(fn.apply(this, arguments)); }; }

function _AsyncGenerator(gen) { var front, back; function send(key, arg) { return new Promise(function (resolve, reject) { var request = { key: key, arg: arg, resolve: resolve, reject: reject, next: null }; if (back) { back = back.next = request; } else { front = back = request; resume(key, arg); } }); } function resume(key, arg) { try { var result = gen[key](arg); var value = result.value; var wrappedAwait = value instanceof _AwaitValue; Promise.resolve(wrappedAwait ? value.wrapped : value).then(function (arg) { if (wrappedAwait) { resume(key === "return" ? "return" : "next", arg); return; } settle(result.done ? "return" : "normal", arg); }, function (err) { resume("throw", err); }); } catch (err) { settle("throw", err); } } function settle(type, value) { switch (type) { case "return": front.resolve({ value: value, done: true }); break; case "throw": front.reject(value); break; default: front.resolve({ value: value, done: false }); break; } front = front.next; if (front) { resume(front.key, front.arg); } else { back = null; } } this._invoke = send; if (typeof gen.return !== "function") { this.return = undefined; } }

if (typeof Symbol === "function" && Symbol.asyncIterator) { _AsyncGenerator.prototype[Symbol.asyncIterator] = function () { return this; }; }

_AsyncGenerator.prototype.next = function (arg) { return this._invoke("next", arg); };

_AsyncGenerator.prototype.throw = function (arg) { return this._invoke("throw", arg); };

_AsyncGenerator.prototype.return = function (arg) { return this._invoke("return", arg); };

function _AwaitValue(value) { this.wrapped = value; }

let DependencyType;
exports.DependencyType = DependencyType;

(function (DependencyType) {
  DependencyType["Dependencies"] = "dependencies";
  DependencyType["DevDependencies"] = "devDependencies";
  DependencyType["PeerDependencies"] = "peerDependencies";
})(DependencyType || (exports.DependencyType = DependencyType = {}));

const DEPENDENCY_TYPES = [DependencyType.Dependencies, DependencyType.DevDependencies, DependencyType.PeerDependencies];

function extractErrorImpl(value) {
  if (value instanceof _tauProlog.default.type.Num) return value.value;

  if (value instanceof _tauProlog.default.type.Term) {
    if (value.args.length === 0) return value.id;

    switch (value.indicator) {
      case `throw/1`:
        return extractErrorImpl(value.args[0]);

      case `error/1`:
        return extractErrorImpl(value.args[0]);

      case `error/2`:
        return Object.assign(extractErrorImpl(value.args[0]), ...extractErrorImpl(value.args[1]));

      case `syntax_error/1`:
        return new _core.ReportError(_core.MessageName.PROLOG_SYNTAX_ERROR, `Syntax error: ${extractErrorImpl(value.args[0])}`);

      case `existence_error/2`:
        return new _core.ReportError(_core.MessageName.PROLOG_EXISTENCE_ERROR, `Existence error: ${extractErrorImpl(value.args[0])} ${extractErrorImpl(value.args[1])} not found`);

      case `line/1`:
        return {
          line: extractErrorImpl(value.args[0])
        };

      case `column/1`:
        return {
          column: extractErrorImpl(value.args[0])
        };

      case `found/1`:
        return {
          found: extractErrorImpl(value.args[0])
        };

      case `./2`:
        return [extractErrorImpl(value.args[0])].concat(extractErrorImpl(value.args[1]));

      case `//2`:
        return `${extractErrorImpl(value.args[0])}/${extractErrorImpl(value.args[1])}`;
    }
  }

  throw `couldn't pretty print because of unsupported node ${value}`;
}

function extractError(val) {
  let err;

  try {
    err = extractErrorImpl(val);
  } catch (caught) {
    if (typeof caught === `string`) {
      throw new _core.ReportError(_core.MessageName.PROLOG_UNKNOWN_ERROR, `Unknown error: ${val} (note: ${caught})`);
    } else {
      throw caught;
    }
  }

  if (typeof err.line !== `undefined` && typeof err.column !== `undefined`) err.message += ` at line ${err.line}, column ${err.column}`;
  return err;
} // Node 8 doesn't have Symbol.asyncIterator
// https://github.com/Microsoft/TypeScript/issues/14151#issuecomment-280812617


if (Symbol.asyncIterator == null) Symbol.asyncIterator = Symbol.for('Symbol.asyncIterator');

class Session {
  constructor(project, source) {
    this.session = void 0;
    this.session = _tauProlog.default.create();
    (0, _tauModule.linkProjectToSession)(this.session, project);
    this.session.consult(source);
  }

  fetchNextAnswer() {
    return new Promise(resolve => {
      this.session.answer(result => {
        resolve(result);
      });
    });
  }

  makeQuery(query) {
    var _this = this;

    return _wrapAsyncGenerator(function* () {
      const parsed = _this.session.query(query);

      if (parsed !== true) throw extractError(parsed);

      while (true) {
        const answer = yield _awaitAsyncGenerator(_this.fetchNextAnswer());
        if (!answer) break;
        if (answer.id === `throw`) throw extractError(answer);
        yield answer;
      }
    })();
  }

}

function parseLink(link) {
  if (link.id === `null`) {
    return null;
  } else {
    return `${link.toJavaScript()}`;
  }
}

function parseLinkToJson(link) {
  if (link.id === `null`) {
    return null;
  } else {
    const val = link.toJavaScript();
    if (typeof val !== `string`) return JSON.stringify(val);

    try {
      return JSON.stringify(JSON.parse(val));
    } catch {
      return JSON.stringify(val);
    }
  }
}

class Constraints {
  static async find(project) {
    return new Constraints(project);
  }

  constructor(project) {
    this.project = void 0;
    this.source = ``;
    this.project = project;

    if (_fslib.xfs.existsSync(_fslib.ppath.join(project.cwd, (0, _fslib.toFilename)(`constraints.pro`)))) {
      this.source = _fslib.xfs.readFileSync(_fslib.ppath.join(project.cwd, (0, _fslib.toFilename)(`constraints.pro`)), `utf8`);
    }
  }

  getProjectDatabase() {
    let database = ``;

    for (const dependencyType of DEPENDENCY_TYPES) database += `dependency_type(${dependencyType}).\n`;

    for (const workspace of this.project.workspacesByCwd.values()) {
      const relativeCwd = workspace.relativeCwd;
      database += `workspace(${escape(relativeCwd)}).\n`;
      database += `workspace_ident(${escape(relativeCwd)}, ${escape(_core.structUtils.stringifyIdent(workspace.locator))}).\n`;
      database += `workspace_version(${escape(relativeCwd)}, ${escape(workspace.manifest.version)}).\n`;

      for (const dependencyType of DEPENDENCY_TYPES) {
        for (const dependency of workspace.manifest[dependencyType].values()) {
          database += `workspace_has_dependency(${escape(relativeCwd)}, ${escape(_core.structUtils.stringifyIdent(dependency))}, ${escape(dependency.range)}, ${dependencyType}).\n`;
        }
      }
    } // Add default never matching predicates to prevent prolog instantiation errors
    // when constraints run in an empty workspace


    database += `workspace(_) :- false.\n`;
    database += `workspace_ident(_, _) :- false.\n`;
    database += `workspace_version(_, _) :- false.\n`; // Add a default never matching predicate to prevent prolog instantiation errors
    // when constraints run in a workspace without dependencies

    database += `workspace_has_dependency(_, _, _, _) :- false.\n`;
    return database;
  }

  getDeclarations() {
    let declarations = ``; // (Cwd, DependencyIdent, DependencyRange, DependencyType)

    declarations += `gen_enforced_dependency(_, _, _, _) :- false.\n`; // (Cwd, Path, Value)

    declarations += `gen_enforced_field(_, _, _) :- false.\n`;
    return declarations;
  }

  get fullSource() {
    return `${this.getProjectDatabase()}\n${this.source}\n${this.getDeclarations()}`;
  }

  createSession() {
    return new Session(this.project, this.fullSource);
  }

  async process() {
    const session = this.createSession();
    return {
      enforcedDependencies: await this.genEnforcedDependencies(session),
      enforcedFields: await this.genEnforcedFields(session)
    };
  }

  async genEnforcedDependencies(session) {
    let enforcedDependencies = [];
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;

    var _iteratorError;

    try {
      for (var _iterator = _asyncIterator(session.makeQuery(`workspace(WorkspaceCwd), dependency_type(DependencyType), gen_enforced_dependency(WorkspaceCwd, DependencyIdent, DependencyRange, DependencyType).`)), _step, _value; _step = await _iterator.next(), _iteratorNormalCompletion = _step.done, _value = await _step.value, !_iteratorNormalCompletion; _iteratorNormalCompletion = true) {
        const answer = _value;

        const workspaceCwd = _fslib.ppath.resolve(this.project.cwd, parseLink(answer.links.WorkspaceCwd));

        const dependencyRawIdent = parseLink(answer.links.DependencyIdent);
        const dependencyRange = parseLink(answer.links.DependencyRange);
        const dependencyType = parseLink(answer.links.DependencyType);
        if (workspaceCwd === null || dependencyRawIdent === null) throw new Error(`Invalid rule`);
        const workspace = this.project.getWorkspaceByCwd(workspaceCwd);

        const dependencyIdent = _core.structUtils.parseIdent(dependencyRawIdent);

        enforcedDependencies.push({
          workspace,
          dependencyIdent,
          dependencyRange,
          dependencyType
        });
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return != null) {
          await _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    return _core.miscUtils.sortMap(enforcedDependencies, [({
      dependencyRange
    }) => dependencyRange !== null ? `0` : `1`, ({
      workspace
    }) => _core.structUtils.stringifyIdent(workspace.locator), ({
      dependencyIdent
    }) => _core.structUtils.stringifyIdent(dependencyIdent)]);
  }

  async genEnforcedFields(session) {
    let enforcedFields = [];
    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;

    var _iteratorError2;

    try {
      for (var _iterator2 = _asyncIterator(session.makeQuery(`workspace(WorkspaceCwd), gen_enforced_field(WorkspaceCwd, FieldPath, FieldValue).`)), _step2, _value2; _step2 = await _iterator2.next(), _iteratorNormalCompletion2 = _step2.done, _value2 = await _step2.value, !_iteratorNormalCompletion2; _iteratorNormalCompletion2 = true) {
        const answer = _value2;

        const workspaceCwd = _fslib.ppath.resolve(this.project.cwd, parseLink(answer.links.WorkspaceCwd));

        const fieldPath = parseLink(answer.links.FieldPath);
        let fieldValue = parseLinkToJson(answer.links.FieldValue);
        if (workspaceCwd === null || fieldPath === null) throw new Error(`Invalid rule`);
        const workspace = this.project.getWorkspaceByCwd(workspaceCwd);
        enforcedFields.push({
          workspace,
          fieldPath,
          fieldValue
        });
      }
    } catch (err) {
      _didIteratorError2 = true;
      _iteratorError2 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
          await _iterator2.return();
        }
      } finally {
        if (_didIteratorError2) {
          throw _iteratorError2;
        }
      }
    }

    return _core.miscUtils.sortMap(enforcedFields, [({
      workspace
    }) => _core.structUtils.stringifyIdent(workspace.locator), ({
      fieldPath
    }) => fieldPath]);
  }

  query(query) {
    var _this2 = this;

    return _wrapAsyncGenerator(function* () {
      const session = _this2.createSession();

      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;

      var _iteratorError3;

      try {
        for (var _iterator3 = _asyncIterator(session.makeQuery(query)), _step3, _value3; _step3 = yield _awaitAsyncGenerator(_iterator3.next()), _iteratorNormalCompletion3 = _step3.done, _value3 = yield _awaitAsyncGenerator(_step3.value), !_iteratorNormalCompletion3; _iteratorNormalCompletion3 = true) {
          const answer = _value3;
          const parsedLinks = {};

          for (const [variable, value] of Object.entries(answer.links)) {
            if (variable !== `_`) {
              parsedLinks[variable] = parseLink(value);
            }
          }

          yield parsedLinks;
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
            yield _awaitAsyncGenerator(_iterator3.return());
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }
    })();
  }

}

exports.Constraints = Constraints;

function escape(what) {
  if (typeof what === `string`) {
    return `'${what}'`;
  } else {
    return `[]`;
  }
}