"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _cli = require("@yarnpkg/cli");

var _core = require("@yarnpkg/core");

var _clipanion = require("clipanion");

var _get = _interopRequireDefault(require("lodash/get"));

var _set = _interopRequireDefault(require("lodash/set"));

var _unset = _interopRequireDefault(require("lodash/unset"));

var _Constraints = require("../Constraints");

var _dec, _dec2, _class, _descriptor, _class2, _temp;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _initializerDefineProperty(target, property, descriptor, context) { if (!descriptor) return; Object.defineProperty(target, property, { enumerable: descriptor.enumerable, configurable: descriptor.configurable, writable: descriptor.writable, value: descriptor.initializer ? descriptor.initializer.call(context) : void 0 }); }

function _applyDecoratedDescriptor(target, property, decorators, descriptor, context) { var desc = {}; Object.keys(descriptor).forEach(function (key) { desc[key] = descriptor[key]; }); desc.enumerable = !!desc.enumerable; desc.configurable = !!desc.configurable; if ('value' in desc || desc.initializer) { desc.writable = true; } desc = decorators.slice().reverse().reduce(function (desc, decorator) { return decorator(target, property, desc) || desc; }, desc); if (context && desc.initializer !== void 0) { desc.value = desc.initializer ? desc.initializer.call(context) : void 0; desc.initializer = undefined; } if (desc.initializer === void 0) { Object.defineProperty(target, property, desc); desc = null; } return desc; }

function _initializerWarningHelper(descriptor, context) { throw new Error('Decorating class property failed. Please ensure that ' + 'proposal-class-properties is enabled and runs after the decorators transform.'); }

// eslint-disable-next-line arca/no-default-export
let ConstraintsCheckCommand = (_dec = _clipanion.Command.Boolean(`--fix`), _dec2 = _clipanion.Command.Path(`constraints`), (_class = (_temp = _class2 = class ConstraintsCheckCommand extends _cli.BaseCommand {
  constructor(...args) {
    super(...args);

    _initializerDefineProperty(this, "fix", _descriptor, this);
  }

  async execute() {
    for (let t = 0; t < 10; ++t) {
      const configuration = await _core.Configuration.find(this.context.cwd, this.context.plugins);
      const {
        project
      } = await _core.Project.find(configuration, this.context.cwd);
      const constraints = await _Constraints.Constraints.find(project);
      const toSave = new Set();
      const report = await _core.StreamReport.start({
        configuration,
        stdout: this.context.stdout
      }, async report => {
        const result = await constraints.process();
        await processDependencyConstraints(toSave, result.enforcedDependencies, {
          fix: this.fix,
          configuration,
          report
        });
        await processFieldConstraints(toSave, result.enforcedFields, {
          fix: this.fix,
          configuration,
          report
        });

        for (const workspace of toSave) {
          workspace.persistManifest();
        }
      });
      if (report.hasErrors()) return report.exitCode();

      if (!this.fix || toSave.size > 0) {
        return 0;
      }
    }

    return 1;
  }

}, _class2.usage = _clipanion.Command.Usage({
  category: `Constraints-related commands`,
  description: `check that the project constraints are met`,
  details: `
      This command will run constraints on your project and emit errors for each one that is found but isn't met. If any error is emitted the process will exit with a non-zero exit code.

      If the \`--fix\` flag is used, Yarn will attempt to automatically fix the issues the best it can, following a multi-pass process (with a maximum of 10 iterations). Some ambiguous patterns cannot be autofixed, in which case you'll have to manually specify the right resolution.

      For more information as to how to write constraints, please consult our dedicated page on our website: https://next.yarnpkg.com/features/constraints.
    `,
  examples: [[`Check that all constraints are satisfied`, `yarn constraints`], [`Autofix all unmet constraints`, `yarn constraints --fix`]]
}), _temp), (_descriptor = _applyDecoratedDescriptor(_class.prototype, "fix", [_dec], {
  configurable: true,
  enumerable: true,
  writable: true,
  initializer: function () {
    return false;
  }
}), _applyDecoratedDescriptor(_class.prototype, "execute", [_dec2], Object.getOwnPropertyDescriptor(_class.prototype, "execute"), _class.prototype)), _class));
exports.default = ConstraintsCheckCommand;

async function processDependencyConstraints(toSave, enforcedDependencies, {
  configuration,
  fix,
  report
}) {
  let hasFixes = false;
  const allIdents = new Map();
  const byWorkspaces = new Map();

  for (const {
    workspace,
    dependencyIdent,
    dependencyRange,
    dependencyType
  } of enforcedDependencies) {
    let byWorkspacesStore = byWorkspaces.get(workspace);
    if (typeof byWorkspacesStore === `undefined`) byWorkspaces.set(workspace, byWorkspacesStore = new Map());
    let byIdentStore = byWorkspacesStore.get(dependencyIdent.identHash);
    if (typeof byIdentStore === `undefined`) byWorkspacesStore.set(dependencyIdent.identHash, byIdentStore = new Map());
    let byDependencyTypeStore = byIdentStore.get(dependencyType);
    if (typeof byDependencyTypeStore === `undefined`) byIdentStore.set(dependencyType, byDependencyTypeStore = new Set());
    allIdents.set(dependencyIdent.identHash, dependencyIdent);
    byDependencyTypeStore.add(dependencyRange);
  }

  for (const [workspace, byWorkspacesStore] of byWorkspaces) {
    for (const [identHash, byIdentStore] of byWorkspacesStore) {
      const dependencyIdent = allIdents.get(identHash);
      if (typeof dependencyIdent === `undefined`) throw new Error(`Assertion failed: The ident should have been registered`);

      for (const [dependencyType, byDependencyTypeStore] of byIdentStore) {
        const expectedRanges = [...byDependencyTypeStore];

        if (expectedRanges.length > 2) {
          report.reportError(_core.MessageName.CONSTRAINTS_AMBIGUITY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${_core.structUtils.prettyIdent(configuration, dependencyIdent)} via conflicting ranges ${expectedRanges.slice(0, -1).map(expectedRange => _core.structUtils.prettyRange(configuration, String(expectedRange))).join(`, `)}, and ${_core.structUtils.prettyRange(configuration, String(expectedRanges[expectedRanges.length - 1]))} (in ${dependencyType})`);
        } else if (expectedRanges.length > 1) {
          report.reportError(_core.MessageName.CONSTRAINTS_AMBIGUITY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${_core.structUtils.prettyIdent(configuration, dependencyIdent)} via conflicting ranges ${_core.structUtils.prettyRange(configuration, String(expectedRanges[0]))} and ${_core.structUtils.prettyRange(configuration, String(expectedRanges[1]))} (in ${dependencyType})`);
        } else {
          const dependencyDescriptor = workspace.manifest[dependencyType].get(dependencyIdent.identHash);
          const [expectedRange] = expectedRanges;

          if (expectedRange !== null) {
            if (!dependencyDescriptor) {
              if (fix) {
                workspace.manifest[dependencyType].set(dependencyIdent.identHash, _core.structUtils.makeDescriptor(dependencyIdent, expectedRange));
                toSave.add(workspace);
                hasFixes = true;
              } else {
                report.reportError(_core.MessageName.CONSTRAINTS_MISSING_DEPENDENCY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${_core.structUtils.prettyIdent(configuration, dependencyIdent)} (via ${_core.structUtils.prettyRange(configuration, expectedRange)}), but doesn't (in ${dependencyType})`);
              }
            } else if (dependencyDescriptor.range !== expectedRange) {
              if (fix) {
                workspace.manifest[dependencyType].set(dependencyIdent.identHash, _core.structUtils.makeDescriptor(dependencyIdent, expectedRange));
                toSave.add(workspace);
                hasFixes = true;
              } else {
                report.reportError(_core.MessageName.CONSTRAINTS_INCOMPATIBLE_DEPENDENCY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${_core.structUtils.prettyIdent(configuration, dependencyIdent)} via ${_core.structUtils.prettyRange(configuration, expectedRange)}, but uses ${_core.structUtils.prettyRange(configuration, dependencyDescriptor.range)} instead (in ${dependencyType})`);
              }
            }
          } else {
            if (dependencyDescriptor) {
              if (fix) {
                workspace.manifest[dependencyType].delete(dependencyIdent.identHash);
                toSave.add(workspace);
                hasFixes = true;
              } else {
                report.reportError(_core.MessageName.CONSTRAINTS_EXTRANEOUS_DEPENDENCY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} has an extraneous dependency on ${_core.structUtils.prettyIdent(configuration, dependencyIdent)} (in ${dependencyType})`);
              }
            }
          }
        }
      }
    }
  }

  return hasFixes;
}

async function processFieldConstraints(toSave, enforcedFields, {
  configuration,
  fix,
  report
}) {
  let hasFixes = false;
  const byWorkspaces = new Map();

  for (const {
    workspace,
    fieldPath,
    fieldValue
  } of enforcedFields) {
    const byWorkspacesStore = _core.miscUtils.getMapWithDefault(byWorkspaces, workspace);

    const byPathStore = _core.miscUtils.getSetWithDefault(byWorkspacesStore, fieldPath);

    byPathStore.add(fieldValue);
  }

  for (const [workspace, byWorkspacesStore] of byWorkspaces) {
    for (const [fieldPath, byPathStore] of byWorkspacesStore) {
      const expectedValues = [...byPathStore];

      if (expectedValues.length > 2) {
        report.reportError(_core.MessageName.CONSTRAINTS_AMBIGUITY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to conflicting values ${expectedValues.slice(0, -1).map(expectedValue => configuration.format(String(expectedValue), `magenta`)).join(`, `)}, or ${configuration.format(String(expectedValues[expectedValues.length - 1]), `magenta`)}`);
      } else if (expectedValues.length > 1) {
        report.reportError(_core.MessageName.CONSTRAINTS_AMBIGUITY, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to conflicting values ${configuration.format(String(expectedValues[0]), `magenta`)} or ${configuration.format(String(expectedValues[1]), `magenta`)}`);
      } else {
        const actualValue = (0, _get.default)(workspace.manifest.raw, fieldPath);
        const [expectedValue] = expectedValues;

        if (expectedValue !== null) {
          if (actualValue === undefined) {
            if (fix) {
              await setWorkspaceField(workspace, fieldPath, expectedValue);
              toSave.add(workspace);
              hasFixes = true;
            } else {
              report.reportError(_core.MessageName.CONSTRAINTS_MISSING_FIELD, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to ${configuration.format(String(expectedValue), `magenta`)}, but doesn't`);
            }
          } else if (JSON.stringify(actualValue) !== expectedValue) {
            if (fix) {
              await setWorkspaceField(workspace, fieldPath, expectedValue);
              toSave.add(workspace);
              hasFixes = true;
            } else {
              report.reportError(_core.MessageName.CONSTRAINTS_INCOMPATIBLE_FIELD, `${_core.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to ${configuration.format(String(expectedValue), `magenta`)}, but is set to ${configuration.format(JSON.stringify(actualValue), `magenta`)} instead`);
            }
          }
        } else {
          if (actualValue !== undefined && actualValue !== null) {
            if (fix) {
              await setWorkspaceField(workspace, fieldPath, null);
              toSave.add(workspace);
              hasFixes = true;
            } else {
              report.reportError(_core.MessageName.CONSTRAINTS_EXTRANEOUS_FIELD, `${_core.structUtils.prettyWorkspace(configuration, workspace)} has an extraneous field ${configuration.format(fieldPath, `cyan`)} set to ${configuration.format(String(expectedValue), `magenta`)}`);
            }
          }
        }
      }
    }
  }

  return hasFixes;
}

async function setWorkspaceField(workspace, fieldPath, value) {
  if (value === null) {
    (0, _unset.default)(workspace.manifest.raw, fieldPath);
  } else {
    (0, _set.default)(workspace.manifest.raw, fieldPath, JSON.parse(value));
  }
}