"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _cli = require("@yarnpkg/cli");

var _core = require("@yarnpkg/core");

var _clipanion = require("clipanion");

var _Constraints = require("../../Constraints");

var _dec, _dec2, _dec3, _class, _descriptor, _descriptor2, _class2, _temp;

function _initializerDefineProperty(target, property, descriptor, context) { if (!descriptor) return; Object.defineProperty(target, property, { enumerable: descriptor.enumerable, configurable: descriptor.configurable, writable: descriptor.writable, value: descriptor.initializer ? descriptor.initializer.call(context) : void 0 }); }

function _applyDecoratedDescriptor(target, property, decorators, descriptor, context) { var desc = {}; Object.keys(descriptor).forEach(function (key) { desc[key] = descriptor[key]; }); desc.enumerable = !!desc.enumerable; desc.configurable = !!desc.configurable; if ('value' in desc || desc.initializer) { desc.writable = true; } desc = decorators.slice().reverse().reduce(function (desc, decorator) { return decorator(target, property, desc) || desc; }, desc); if (context && desc.initializer !== void 0) { desc.value = desc.initializer ? desc.initializer.call(context) : void 0; desc.initializer = undefined; } if (desc.initializer === void 0) { Object.defineProperty(target, property, desc); desc = null; } return desc; }

function _initializerWarningHelper(descriptor, context) { throw new Error('Decorating class property failed. Please ensure that ' + 'proposal-class-properties is enabled and runs after the decorators transform.'); }

function _asyncIterator(iterable) { var method; if (typeof Symbol !== "undefined") { if (Symbol.asyncIterator) { method = iterable[Symbol.asyncIterator]; if (method != null) return method.call(iterable); } if (Symbol.iterator) { method = iterable[Symbol.iterator]; if (method != null) return method.call(iterable); } } throw new TypeError("Object is not async iterable"); }

// eslint-disable-next-line arca/no-default-export
let ConstraintsQueryCommand = (_dec = _clipanion.Command.Boolean(`--json`), _dec2 = _clipanion.Command.String(), _dec3 = _clipanion.Command.Path(`constraints`, `query`), (_class = (_temp = _class2 = class ConstraintsQueryCommand extends _cli.BaseCommand {
  constructor(...args) {
    super(...args);

    _initializerDefineProperty(this, "json", _descriptor, this);

    _initializerDefineProperty(this, "query", _descriptor2, this);
  }

  async execute() {
    const configuration = await _core.Configuration.find(this.context.cwd, this.context.plugins);
    const {
      project
    } = await _core.Project.find(configuration, this.context.cwd);
    const constraints = await _Constraints.Constraints.find(project);
    let query = this.query;
    if (!query.endsWith(`.`)) query = `${query}.`;
    const report = await _core.StreamReport.start({
      configuration,
      json: this.json,
      stdout: this.context.stdout
    }, async report => {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;

      var _iteratorError;

      try {
        for (var _iterator = _asyncIterator(constraints.query(query)), _step, _value; _step = await _iterator.next(), _iteratorNormalCompletion = _step.done, _value = await _step.value, !_iteratorNormalCompletion; _iteratorNormalCompletion = true) {
          const result = _value;
          const lines = Array.from(Object.entries(result));
          const lineCount = lines.length;
          const maxVariableNameLength = lines.reduce((max, [variableName]) => Math.max(max, variableName.length), 0);

          for (let i = 0; i < lineCount; i++) {
            const [variableName, value] = lines[i];
            report.reportInfo(null, `${getLinePrefix(i, lineCount)}${variableName.padEnd(maxVariableNameLength, ` `)} = ${valueToString(value)}`);
          }

          report.reportJson(result);
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return != null) {
            await _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    });
    return report.exitCode();
  }

}, _class2.usage = _clipanion.Command.Usage({
  category: `Constraints-related commands`,
  description: `query the constraints fact database`,
  details: `
      This command will output all matches to the given prolog query.

      If the \`--json\` flag is set the output will follow a JSON-stream output also known as NDJSON (https://github.com/ndjson/ndjson-spec).
    `,
  examples: [[`List all dependencies throughout the workspace`, `yarn constraints query 'workspace_has_dependency(_, DependencyName, _, _).'`]]
}), _temp), (_descriptor = _applyDecoratedDescriptor(_class.prototype, "json", [_dec], {
  configurable: true,
  enumerable: true,
  writable: true,
  initializer: function () {
    return false;
  }
}), _descriptor2 = _applyDecoratedDescriptor(_class.prototype, "query", [_dec2], {
  configurable: true,
  enumerable: true,
  writable: true,
  initializer: null
}), _applyDecoratedDescriptor(_class.prototype, "execute", [_dec3], Object.getOwnPropertyDescriptor(_class.prototype, "execute"), _class.prototype)), _class));
exports.default = ConstraintsQueryCommand;

function valueToString(value) {
  if (typeof value !== `string`) return `${value}`;
  if (value.match(/^[a-zA-Z][a-zA-Z0-9_]+$/)) return value;
  return `'${value}'`;
}

function getLinePrefix(index, count) {
  const isFirst = index === 0;
  const isLast = index === count - 1;
  if (isFirst && isLast) return ``;
  if (isFirst) return `┌ `;
  if (isLast) return `└ `;
  return `│ `;
}